package ru.dbotthepony.mc.otm.storage;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.IStorageGridCell;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.*;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class StorageGrid {
	public final StorageItemView global_item_view = new StorageItemView();
	private final HashSet<IStorageGridCell> cells = new HashSet<>();

	private final HashMap<Class<? extends IStorageObject>, IStorageView<? extends IStorageObject>> views = new HashMap<>();
	private final HashMap<Class<? extends IStorageObject>, ArrayList<IStorageConsumer<? extends IStorageObject>>> consumers = new HashMap<>();
	private final HashMap<Class<? extends IStorageObject>, IStorageComponent<? extends IStorageObject>> virtual_components = new HashMap<>();

	public StorageGrid() {
		views.put(ItemStackObject.class, global_item_view);
	}

	@SuppressWarnings("unchecked")
	public <T extends IStorageObject> T insertObject(Class<T> type, T object, boolean simulate) {
		var leftover = object;

		for (var consumer : consumers.computeIfAbsent(type, (k) -> new ArrayList<>())) {
			leftover = ((IStorageConsumer<T>) consumer).insertObject(leftover, simulate);

			if (leftover.isEmpty()) {
				return leftover;
			}
		}

		return leftover;
	}

	/**
	 * @param type Class, representing identity of stored object
	 * @param <T> The identity itself
	 * @return a virtual IStorageComponent<T>, or null if storage grid contain no view for provided type
	 */
	@SuppressWarnings("unchecked")
	@Nullable
	public <T extends IStorageObject> IStorageComponent<T> getVirtualComponent(final Class<T> type) {
		return (IStorageComponent<T>) virtual_components.computeIfAbsent(type, (k) -> {
			final var get_view = (IStorageView<T>) views.get(type);

			if (get_view == null)
				return null;

			return new IStorageComponent<T>() {
				@Override
				public T insertObject(T obj, boolean simulate) {
					return StorageGrid.this.insertObject(type, obj, simulate);
				}

				@Override
				public T getStoredObject(UUID id) {
					return get_view.getStoredObject(id);
				}

				@Override
				public T extractObject(UUID id, BigDecimal amount, boolean simulate) {
					return get_view.extractObject(id, amount, simulate);
				}

				@Override
				public List<IStorageTuple<T>> getStorageObjects() {
					return get_view.getStorageObjects();
				}

				@Override
				public Class<T> storageIdentity() {
					return type;
				}

				@Override
				public boolean addListener(IStorageListener<T> listener) {
					return get_view.addListener(listener);
				}

				@Override
				public boolean removeListener(IStorageListener<T> listener) {
					return get_view.removeListener(listener);
				}
			};
		});
	}

	public int size() {
		return cells.size();
	}

	@SuppressWarnings("unchecked")
	public <T extends IStorageObject> void add(IStorageIdentity<T> identity) {
		final var view = (IStorageView<T>) views.get(identity.storageIdentity());

		if (view != null && identity instanceof IStorageView provider) {
			((IStorageView<T>) provider).addListenerAuto((IStorageListener<T>) view);
		} else if (view != null && identity instanceof IStorageTrigger provider) {
			((IStorageTrigger<T>) provider).addListener((IStorageListener<T>) view);
		}

		if (identity instanceof IStorageConsumer provider) {
			consumers.computeIfAbsent(identity.storageIdentity(), (k) -> new ArrayList<>()).add(provider);
		}
	}

	public boolean add(LazyOptional<IStorageGridCell> cell) {
		if (!cell.isPresent())
			return false;

		var resolved = cell.resolve();

		if (resolved.isPresent()) {
			final var value = resolved.get();

			if (add(value)) {
				cell.addListener((lazy) -> remove(value));
				return true;
			}
		}

		return false;
	}

	public boolean add(IStorageGridCell cell) {
		if (cells.add(cell)) {
			cell.setStorageGrid(this);

			for (var identity : cell.getComponents()) {
				add(identity);
			}

			return true;
		}

		return false;
	}

	@SuppressWarnings("unchecked")
	public <T extends IStorageObject> void remove(IStorageIdentity<T> identity) {
		final var view = (IStorageView<T>) views.get(identity.storageIdentity());

		if (view != null && identity instanceof IStorageView provider) {
			((IStorageView<T>) provider).removeListenerAuto((IStorageListener<T>) view);
		} else if (view != null && identity instanceof IStorageTrigger provider) {
			((IStorageTrigger<T>) provider).removeListener((IStorageListener<T>) view);
		}

		if (identity instanceof IStorageConsumer provider) {
			consumers.computeIfAbsent(identity.storageIdentity(), (k) -> new ArrayList<>()).remove(provider);
		}
	}

	public boolean remove(IStorageGridCell cell) {
		if (cells.remove(cell)) {
			cell.setStorageGrid(null);

			for (var identity : cell.getComponents()) {
				remove(identity);
			}

			return true;
		}

		return false;
	}

	public void detach(IStorageGridCell cell) {
		if (cells.contains(cell)) {
			for (var identity : cell.getComponents()) {
				remove(identity);
			}
		}
	}

	public void attach(IStorageGridCell cell) {
		if (cells.contains(cell)) {
			for (var identity : cell.getComponents()) {
				add(identity);
			}
		}
	}

	public StorageGrid merge(StorageGrid other) {
		if (other == this)
			return this;

		if (size() < other.size()) {
			var copy = List.copyOf(cells);

			for (var cell : copy) {
				remove(cell);
				other.add(cell);
			}

			return other;
		} else {
			var copy = List.copyOf(other.cells);

			for (var cell : copy) {
				other.remove(cell);
				add(cell);
			}

			return this;
		}
	}

	public static void scheduleDiscoverNeighbours(IStorageGridCell cell, Level level, BlockPos pos) {
		OverdriveThatMatters.tickUntil(level, () -> !cell.isValidStorageCell() || cell.connectOrCreateStorageGrid(level, pos, true));
	}
}
