package ru.dbotthepony.mc.otm.storage;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;

import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.*;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class StorageItemView implements IStorageListener<ItemStackObject>, IStorageView<ItemStackObject> {
	public record LocalTuple(ItemStackObject object, UUID id, List<RemoteTuple> tuples) implements IStorageTuple<ItemStackObject> {
		public static final LocalTuple EMPTY = new LocalTuple(ItemStackObject.EMPTY, new UUID(0, 0), List.of());

		@Override
		public boolean equals(Object obj) {
			return obj instanceof LocalTuple tuple && tuple.id.equals(id) || obj instanceof UUID id && this.id.equals(id);
		}

		@Override
		public int hashCode() {
			return id.hashCode();
		}
	}

	@Override
	public Class<ItemStackObject> storageIdentity() {
		return ItemStackObject.class;
	}

	public record RemoteTuple(ItemStack stack, UUID remote_id, IStorageView<ItemStackObject> provider, LocalTuple local) {
		public static final RemoteTuple EMPTY = new RemoteTuple(ItemStack.EMPTY, new UUID(0, 0), new IStorageView<>() {
			@Override
			public Class<ItemStackObject> storageIdentity() {
				return ItemStackObject.class;
			}

			@Override
			public boolean addListener(IStorageListener<ItemStackObject> listener) {
				return false;
			}

			@Override
			public boolean removeListener(IStorageListener<ItemStackObject> listener) {
				return false;
			}

			@Override
			public ItemStackObject getStoredObject(UUID id) {
				return ItemStackObject.EMPTY;
			}

			@Override
			public ItemStackObject extractObject(UUID id, BigDecimal amount, boolean simulate) {
				return ItemStackObject.EMPTY;
			}

			@Override
			public List<IStorageTuple<ItemStackObject>> getStorageObjects() {
				return List.of();
			}
		}, LocalTuple.EMPTY);

		public ItemStackObject extract(BigDecimal amount, boolean simulate) {
			return provider.extractObject(remote_id, amount, simulate);
		}

		public BigDecimal extractCount(BigDecimal amount, boolean simulate) {
			return provider.extractObjectCount(remote_id, amount, simulate);
		}

		@Override
		public boolean equals(Object obj) {
			return obj instanceof RemoteTuple tuple && tuple.remote_id.equals(remote_id) || obj instanceof UUID id && id.equals(remote_id);
		}

		@Override
		public int hashCode() {
			return remote_id.hashCode();
		}
	}

	// удаленный UUID -> Кортеж
	private final HashMap<UUID, RemoteTuple> indexed_by_remote_uuid = new HashMap<>();

	// локальный UUID -> Локальный кортеж
	private final HashMap<UUID, LocalTuple> indexed_by_local_uuid = new HashMap<>();

	// Item -> Список Локальных кортежей
	private final HashMap<Item, ArrayList<LocalTuple>> indexed_by_item = new HashMap<>();

	public List<ItemStack> getCopyRawItemList() {
		final var list = new ArrayList<ItemStack>(indexed_by_remote_uuid.size());

		for (var state : indexed_by_remote_uuid.values())
			list.add(state.stack.copy());

		return list;
	}

	public List<ItemStack> getRawItemList() {
		final var list = new ArrayList<ItemStack>(indexed_by_remote_uuid.size());

		for (var state : indexed_by_remote_uuid.values())
			list.add(state.stack);

		return list;
	}

	@Override
	public List<IStorageTuple<ItemStackObject>> getStorageObjects() {
		int capacity = 0;

		for (var list : indexed_by_item.values())
			capacity += list.size();

		final var output = new ArrayList<IStorageTuple<ItemStackObject>>(capacity);

		for (var listing : indexed_by_item.values())
			output.addAll(listing);

		return output;
	}

	public List<ItemStack> getCombinedItemList() {
		int capacity = 0;

		for (var list : indexed_by_item.values())
			capacity += list.size();

		final var output = new ArrayList<ItemStack>(capacity);

		for (var listing : indexed_by_item.values())
			for (var combined : listing)
				output.add(combined.object.stack());

		return output;
	}

	@Override
	public void addObject(ItemStackObject _stack, UUID id, IStorageView<ItemStackObject> provider) {
		if (indexed_by_remote_uuid.containsKey(id)) {
			throw new IllegalStateException("Already tracking tuple with id " + id);
		}

		var stack = _stack.stack();

		var items = indexed_by_item.computeIfAbsent(stack.getItem(), (k) -> new ArrayList<>());
		LocalTuple local_tuple = null;

		for (var item : items) {
			if (ItemStack.tagMatches(item.object.stack(), stack)) {
				item.object.stack().grow(stack.getCount());
				local_tuple = item;
				break;
			}
		}

		boolean added = local_tuple == null;

		if (added) {
			local_tuple = new LocalTuple((ItemStackObject) _stack.copy(), UUID.randomUUID(), new ArrayList<>(1));
			items.add(local_tuple);
			indexed_by_local_uuid.put(local_tuple.id, local_tuple);
		}

		var tuple = new RemoteTuple(stack.copy(), id, provider, local_tuple);
		local_tuple.tuples.add(tuple);
		indexed_by_remote_uuid.put(id, tuple);

		if (added) {
			for (var listener : listeners) {
				listener.addObject(local_tuple.object, local_tuple.id, this);
			}
		} else {
			for (var listener : listeners) {
				listener.changeObject(local_tuple.id, local_tuple.object.getCount());
			}
		}
	}

	@Override
	public void changeObject(UUID id, BigDecimal new_count) {
		assert new_count.compareTo(BigDecimal.ZERO) > 0;
		var tuple = indexed_by_remote_uuid.get(id);

		if (tuple == null)
			throw new IllegalStateException("No such tuple with id " + id);

		var delta = new_count.intValue() - tuple.stack.getCount();
		tuple.stack.setCount(new_count.intValue());
		tuple.local.object.stack().grow(delta);

		for (var listener : listeners) {
			listener.changeObject(tuple.local.id, tuple.local.object.getCount());
		}
	}

	@Override
	public void removeObject(UUID id) {
		var tuple = indexed_by_remote_uuid.get(id);

		if (tuple == null)
			throw new IllegalStateException("No such tuple with id " + id);

		var item = tuple.local.object.stack().getItem();
		tuple.local.object.stack().shrink(tuple.stack.getCount());
		tuple.local.tuples.remove(tuple);

		indexed_by_remote_uuid.remove(id);

		if (tuple.local.object.stack().getCount() <= 0 || tuple.local.tuples.size() == 0) {
			assert tuple.local.object.stack().getCount() == tuple.local.tuples.size();

			indexed_by_local_uuid.remove(tuple.local.id);
			indexed_by_item.get(item).remove(tuple.local);

			for (var listener : listeners) {
				listener.removeObject(tuple.local.id);
			}
		}
	}

	private final Set<IStorageListener<ItemStackObject>> listeners = new HashSet<>();

	@Override
	public boolean addListener(IStorageListener<ItemStackObject> listener) {
		return listeners.add(listener);
	}

	@Override
	public boolean removeListener(IStorageListener<ItemStackObject> listener) {
		return listeners.remove(listener);
	}

	@Override
	public ItemStackObject getStoredObject(UUID id) {
		return indexed_by_local_uuid.getOrDefault(id, LocalTuple.EMPTY).object;
	}

	@Override
	public ItemStackObject extractObject(UUID id, BigDecimal _amount, boolean simulate) {
		var tuple = indexed_by_local_uuid.get(id);

		if (tuple == null)
			return ItemStackObject.EMPTY;

		int amount = _amount.intValue();

		if (amount == 0)
			return ItemStackObject.EMPTY;

		if (amount == -1)
			amount = tuple.object.stack().getMaxStackSize();

		int extract = Math.min(amount, tuple.object.stack().getCount());
		int extracted = 0;

		var copy = tuple.object.stack().copy();

		for (var remote_tuple : tuple.tuples) {
			extracted += remote_tuple.extractCount(new BigDecimal(extract - extracted), simulate).intValue();

			if (extracted >= extract)
				break;
		}

		if (extracted > 0) {
			copy.setCount(extracted);
			return new ItemStackObject(copy);
		}

		return ItemStackObject.EMPTY;
	}
}
